@@ -193,9 +193,9 @@ static int getToken()
                     break;
                 case 'f':
                     /*
-                                  QTranslator::findMessage() has the same parameters as
-                                  QApplication::translate().
-                                   */
+                     * QTranslator::findMessage() has the same parameters as
+                     * QApplication::translate().
+                     */
                     if ( strcmp(yyIdent + 1, "indMessage") == 0 )
                         return Tok_translate;
                     break;
@@ -359,9 +359,12 @@ static int getToken()
                     yyString[yyStringLen] = '\0';
 
                     if ( yyCh != quoteChar ) {
-                        printf("%c\n", yyCh);
-                        qWarning( "%s:%d: Unterminated string",
-                          (const char *) yyFileName, yyLineNo );
+                        if (trippelQuote)
+                            qWarning("%s:%d: Empty or unterminated triple quoted string",
+                                    (const char *)yyFileName, yyLineNo);
+                        else
+                            qWarning("%s:%d: Unterminated string",
+                                    (const char *)yyFileName, yyLineNo);
                     }
 
                     if ( yyCh == EOF ) {
