@@ -1,3 +1,16 @@
+/**********************************************************************
+** Copyright (C) 2002-2007 Detlev Offenbach <detlev@die-offenbachs.de>
+**
+** This is a modified version of lupdate. The original is part of Qt-Linguist.
+** The copyright of the original file can be found below.
+**
+** This version is modified to handle python sources.
+**
+**   The file is provided AS IS with NO WARRANTY OF ANY KIND,
+**   INCLUDING THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR
+**   A PARTICULAR PURPOSE.
+****************************************************************************/
+
 /****************************************************************************
 **
 ** Copyright (C) 1992-2005 Trolltech AS. All rights reserved.
@@ -21,13 +34,13 @@
 
 #include "metatranslator.h"
 
-#include <qapplication.h>
+#include <qcoreapplication.h>
 #include <qbytearray.h>
 #include <qfile.h>
-#include <qmessagebox.h>
 #include <qtextcodec.h>
 #include <qtextstream.h>
 #include <qxml.h>
+#include <qlocale.h>
 
 static bool encodingIsUtf8( const QXmlAttributes& atts )
 {
@@ -48,7 +61,7 @@ public:
     TsHandler( MetaTranslator *translator )
         : tor( translator ), type( MetaTranslatorMessage::Finished ),
           inMessage( false ), ferrorCount( 0 ), contextIsUtf8( false ),
-          messageIsUtf8( false ) { }
+          messageIsUtf8( false ), m_isPlural(false) { }
 
     virtual bool startElement( const QString& namespaceURI,
                                const QString& localName, const QString& qName,
@@ -57,20 +70,25 @@ public:
                              const QString& localName, const QString& qName );
     virtual bool characters( const QString& ch );
     virtual bool fatalError( const QXmlParseException& exception );
+    QString language() const { return m_language; }
 
 private:
     MetaTranslator *tor;
     MetaTranslatorMessage::Type type;
     bool inMessage;
+    QString m_language;
     QString context;
     QString source;
     QString comment;
-    QString translation;
+    QStringList translations;
+    QString m_fileName;
+    int     m_lineNumber;
 
     QString accum;
     int ferrorCount;
     bool contextIsUtf8;
     bool messageIsUtf8;
+    bool m_isPlural;
 };
 
 bool TsHandler::startElement( const QString& /* namespaceURI */,
@@ -93,19 +111,28 @@ bool TsHandler::startElement( const QString& /* namespaceURI */,
             }
         }
     } else {
-        if ( qName == QString("context") ) {
+        if ( qName == QString("TS") ) {
+            m_language = atts.value(QLatin1String("language"));
+        } else if ( qName == QString("context") ) {
             context.truncate( 0 );
             source.truncate( 0 );
             comment.truncate( 0 );
-            translation.truncate( 0 );
+            translations.clear();
             contextIsUtf8 = encodingIsUtf8( atts );
         } else if ( qName == QString("message") ) {
             inMessage = true;
             type = MetaTranslatorMessage::Finished;
             source.truncate( 0 );
             comment.truncate( 0 );
-            translation.truncate( 0 );
+            translations.clear();
             messageIsUtf8 = encodingIsUtf8( atts );
+            m_isPlural = atts.value(QLatin1String("numerus")).compare(QLatin1String("yes")) == 0;
+        } else if (qName == QString("location") && inMessage) {
+            bool bOK;
+            int lineNo = atts.value(QString("line")).toInt(&bOK);
+            if (!bOK) lineNo = -1;
+            m_fileName = atts.value(QString("filename"));
+            m_lineNumber = lineNo;
         } else if ( qName == QString("translation") ) {
             for ( int i = 0; i < atts.length(); i++ ) {
                 if ( atts.qName(i) == QString("type") ) {
@@ -141,25 +168,31 @@ bool TsHandler::endElement( const QString& /* namespaceURI */,
             if ( contextIsUtf8 )
                 tor->insert( MetaTranslatorMessage(context.toUtf8(),
                              ContextComment,
-                             accum.toUtf8(), QString(), true,
+                             accum.toUtf8(), QString(), 0,
+                             QStringList(), true,
                              MetaTranslatorMessage::Unfinished) );
             else
                 tor->insert( MetaTranslatorMessage(context.toAscii(),
                              ContextComment,
-                             accum.toAscii(), QString(), false,
+                             accum.toAscii(), QString(), 0,
+                             QStringList(), false,
                              MetaTranslatorMessage::Unfinished) );
         }
+    } else if ( qName == QString("numerusform") ) {
+        translations.append(accum);
+        m_isPlural = true;
     } else if ( qName == QString("translation") ) {
-        translation = accum;
+        if (translations.isEmpty())
+            translations.append(accum);
     } else if ( qName == QString("message") ) {
         if ( messageIsUtf8 )
             tor->insert( MetaTranslatorMessage(context.toUtf8(), source.toUtf8(),
-                                               comment.toUtf8(), translation,
-                                               true, type) );
+                                               comment.toUtf8(), m_fileName, m_lineNumber,
+                                               translations, true, type, m_isPlural) );
         else
             tor->insert( MetaTranslatorMessage(context.toAscii(), source.toAscii(),
-                                               comment.toAscii(), translation,
-                                               false, type) );
+                                               comment.toAscii(), m_fileName, m_lineNumber,
+                                               translations, false, type, m_isPlural) );
         inMessage = false;
     }
     return true;
@@ -180,11 +213,7 @@ bool TsHandler::fatalError( const QXmlParseException& exception )
         msg.sprintf( "Parse error at line %d, column %d (%s).",
                      exception.lineNumber(), exception.columnNumber(),
                      exception.message().toLatin1().data() );
-        if ( qApp == 0 )
-            fprintf( stderr, "XML error: %s\n", msg.toLatin1().data() );
-        else
-            QMessageBox::information(0,
-                                      QObject::tr("Qt Linguist"), msg );
+        fprintf( stderr, "XML error: %s\n", msg.toLatin1().data() );
     }
     return false;
 }
@@ -245,17 +274,19 @@ static QString evilBytes( const QByteArray& str, bool utf8 )
 }
 
 MetaTranslatorMessage::MetaTranslatorMessage()
-    : utfeight( false ), ty( Unfinished )
+    : utfeight( false ), ty( Unfinished ), m_plural(false)
 {
 }
 
 MetaTranslatorMessage::MetaTranslatorMessage( const char *context,
                                               const char *sourceText,
                                               const char *comment,
-                                              const QString& translation,
-                                              bool utf8, Type type )
-    : TranslatorMessage( context, sourceText, comment, translation ),
-      utfeight( false ), ty( type )
+                                              const QString &fileName,
+                                              int lineNumber,
+                                              const QStringList& translations,
+                                              bool utf8, Type type, bool plural )
+    : TranslatorMessage( context, sourceText, comment, fileName, lineNumber, translations ),
+      utfeight( false ), ty( type ), m_plural(plural)
 {
     /*
       Don't use UTF-8 if it makes no difference. UTF-8 should be
@@ -287,7 +318,7 @@ MetaTranslatorMessage::MetaTranslatorMessage( const char *context,
 }
 
 MetaTranslatorMessage::MetaTranslatorMessage( const MetaTranslatorMessage& m )
-    : TranslatorMessage( m ), utfeight( m.utfeight ), ty( m.ty )
+    : TranslatorMessage( m ), utfeight( m.utfeight ), ty( m.ty ), m_plural(m.m_plural)
 {
 }
 
@@ -297,6 +328,7 @@ MetaTranslatorMessage& MetaTranslatorMessage::operator=(
     TranslatorMessage::operator=( m );
     utfeight = m.utfeight;
     ty = m.ty;
+    m_plural = m.m_plural;
     return *this;
 }
 
@@ -310,10 +342,11 @@ bool MetaTranslatorMessage::operator==( const MetaTranslatorMessage& m ) const
 bool MetaTranslatorMessage::operator<( const MetaTranslatorMessage& m ) const
 {
     int delta = qstrcmp( context(), m.context() );
-    if ( delta == 0 )
+    if ( delta == 0 ) {
         delta = qstrcmp( sourceText(), m.sourceText() );
-    if ( delta == 0 )
-        delta = qstrcmp( comment(), m.comment() );
+        if ( delta == 0 )
+            delta = qstrcmp( comment(), m.comment() );
+    }
     return delta < 0;
 }
 
@@ -352,15 +385,17 @@ bool MetaTranslator::load( const QString& filename )
     QXmlSimpleReader reader;
     reader.setFeature( "http://xml.org/sax/features/namespaces", false );
     reader.setFeature( "http://xml.org/sax/features/namespace-prefixes", true );
-    reader.setFeature( "http://trolltech.com/xml/features/report-whitespace"
-                       "-only-CharData", false );
-    QXmlDefaultHandler *hand = new TsHandler( this );
-    reader.setContentHandler( hand );
-    reader.setErrorHandler( hand );
+    TsHandler *hand = new TsHandler( this );
+    reader.setContentHandler( static_cast<QXmlDefaultHandler*>(hand) );
+    reader.setErrorHandler( static_cast<QXmlDefaultHandler*>(hand) );
 
     bool ok = reader.parse( in );
     reader.setContentHandler( 0 );
     reader.setErrorHandler( 0 );
+
+    m_language = hand->language();
+    makeFileNamesAbsolute(QFileInfo(filename).absoluteDir());
+
     delete hand;
     f.close();
     return ok;
@@ -375,7 +410,13 @@ bool MetaTranslator::save( const QString& filename ) const
     QTextStream t( &f );
     t.setCodec( QTextCodec::codecForName("ISO-8859-1") );
 
-    t << "<!DOCTYPE TS><TS>\n";
+    //### The xml prolog allows processors to easily detect the correct encoding
+    t << "<?xml version=\"1.0\"";
+    t << " encoding=\"utf-8\"";
+    t << "?>\n<!DOCTYPE TS><TS version=\"1.1\"";
+    if (!languageCode().isEmpty() && languageCode() != QLatin1String("C"))
+        t << " language=\"" << languageCode() << "\"";
+    t << ">\n";
     if ( codecName != "ISO-8859-1" )
         t << "<defaultcodec>" << codecName << "</defaultcodec>\n";
     TMM::ConstIterator m = mm.begin();
@@ -408,29 +449,49 @@ bool MetaTranslator::save( const QString& filename ) const
               << "</comment>\n";
 
         for ( i = inv.begin(); i != inv.end(); ++i ) {
+            MetaTranslatorMessage msg = *i;
             // no need for such noise
-            if ( (*i).type() == MetaTranslatorMessage::Obsolete &&
-                 (*i).translation().isEmpty() )
+            if ( msg.type() == MetaTranslatorMessage::Obsolete &&
+                 msg.translation().isEmpty() )
                 continue;
 
             t << "    <message";
-            if ( (*i).utf8() )
+            if ( msg.utf8() )
                 t << " encoding=\"UTF-8\"";
-            t << ">\n"
-              << "        <source>" << evilBytes( (*i).sourceText(),
-                                                  (*i).utf8() )
+            if ( msg.isPlural() )
+                t << " numerus=\"yes\"";
+            t << ">\n";
+            if (!msg.fileName().isEmpty() && msg.lineNumber() >= 0) {
+                QDir tsPath = QFileInfo(filename).absoluteDir();
+                QString fn = tsPath.relativeFilePath(msg.fileName()).replace('\\','/');
+                t << "        <location filename=\"" << fn << "\" line=\"" << msg.lineNumber() << "\"/>\n";
+            }
+            t  << "        <source>" << evilBytes( msg.sourceText(),
+                                                  msg.utf8() )
               << "</source>\n";
-            if ( !QByteArray((*i).comment()).isEmpty() )
-                t << "        <comment>" << evilBytes( (*i).comment(),
-                                                       (*i).utf8() )
+            if ( !QByteArray(msg.comment()).isEmpty() )
+                t << "        <comment>" << evilBytes( msg.comment(),
+                                                       msg.utf8() )
                   << "</comment>\n";
             t << "        <translation";
-            if ( (*i).type() == MetaTranslatorMessage::Unfinished )
+            if ( msg.type() == MetaTranslatorMessage::Unfinished )
                 t << " type=\"unfinished\"";
-            else if ( (*i).type() == MetaTranslatorMessage::Obsolete )
+            else if ( msg.type() == MetaTranslatorMessage::Obsolete )
                 t << " type=\"obsolete\"";
-            t << ">" << protect( (*i).translation().toUtf8() )
-              << "</translation>\n";
+            t << ">";
+            if (msg.isPlural()) {
+                t << "\n";
+                QLocale::Language l;
+                QLocale::Country c;
+                languageAndCountry(m_language, &l, &c);
+                QStringList translns = normalizedTranslations(msg, l, c);
+                for (int j = 0; j < qMax(1, translns.count()); ++j)
+                    t << "            <numerusform>" << protect( translns.value(j).toUtf8() ) << "</numerusform>\n";
+                t << "        ";
+            } else {
+                t << protect( msg.translation().toUtf8() );
+            }
+            t << "</translation>\n";
             t << "    </message>\n";
         }
         t << "</context>\n";
@@ -440,6 +501,21 @@ bool MetaTranslator::save( const QString& filename ) const
     return true;
 }
 
+void MetaTranslator::languageAndCountry(const QString &languageCode, QLocale::Language *lang, QLocale::Country *country)
+{
+    QLocale locale(languageCode);
+    if (lang)
+        *lang = locale.language();
+
+    if (country) {
+        if (languageCode.indexOf(QLatin1Char('_')) != -1) {
+            *country = locale.country();
+        } else {
+            *country = QLocale::AnyCountry;
+        }
+    }
+}
+
 bool MetaTranslator::release( const QString& filename, bool verbose,
                               bool ignoreUnfinished,
                               Translator::SaveMode mode ) const
@@ -463,7 +539,7 @@ bool MetaTranslator::release( const QString& filename, bool verbose,
                 QByteArray context = m.key().context();
                 QByteArray sourceText = m.key().sourceText();
                 QByteArray comment = m.key().comment();
-                QString translation = m.key().translation();
+                QStringList translations = m.key().translations();
 
                 if ( !ignoreUnfinished
                      || m.key().type() != MetaTranslatorMessage::Unfinished ) {
@@ -482,7 +558,8 @@ bool MetaTranslator::release( const QString& filename, bool verbose,
                         tor.insert( m.key() );
                     } else {
                         tor.insert( TranslatorMessage(context, sourceText, "",
-                                                       translation) );
+                                                      QString(), -1, translations) );
+                        //filename and lineNumbers will be ignored from now.
                     }
                 }
             }
@@ -498,10 +575,54 @@ bool MetaTranslator::release( const QString& filename, bool verbose,
     return saved;
 }
 
+QString MetaTranslator::languageCode() const
+{
+    return m_language;
+}
+
+void MetaTranslator::setLanguageCode(const QString &languageCode)
+{
+    m_language = languageCode;
+}
+
 bool MetaTranslator::contains( const char *context, const char *sourceText,
                                const char *comment ) const
 {
-    return mm.contains(MetaTranslatorMessage(context, sourceText, comment));
+    return mm.contains(MetaTranslatorMessage(context, sourceText, comment, QString(), 0));
+}
+
+MetaTranslatorMessage MetaTranslator::find( const char *context,
+        const char *sourceText, const char *comment ) const
+{
+    QMap<MetaTranslatorMessage, int>::const_iterator it = 
+        mm.constFind(MetaTranslatorMessage(context, sourceText, comment, QString(), 0));
+    return (it == mm.constEnd() ? MetaTranslatorMessage() : it.key());
+}
+
+MetaTranslatorMessage MetaTranslator::find(const char *context, const char *comment, 
+                                const QString &fileName, int lineNumber) const
+{
+    if (lineNumber >= 0 && !fileName.isEmpty()) {
+        MetaTranslatorMessage m;
+
+        for (QMap<MetaTranslatorMessage, int>::const_iterator it = mm.constBegin();
+            it != mm.constEnd(); ++it) {
+            m = it.key();
+            int delta = qstrcmp(m.context(), context);
+            if (delta == 0) {
+                delta = qstrcmp(m.comment(), comment);
+                if (delta == 0) {
+                    delta = QString::compare(m.fileName(), fileName);
+                    if (delta == 0) {
+                        delta = m.lineNumber() - lineNumber;
+                        if (delta == 0)
+                            return m;
+                    }
+                }
+            }
+        }
+    }
+    return MetaTranslatorMessage();
 }
 
 void MetaTranslator::insert( const MetaTranslatorMessage& m )
@@ -547,6 +668,23 @@ void MetaTranslator::stripEmptyContexts()
     mm = newmm;
 }
 
+void MetaTranslator::makeFileNamesAbsolute(const QDir &oldPath)
+{
+    TMM newmm;
+    for (TMM::iterator m = mm.begin(); m != mm.end(); ++m) {
+        MetaTranslatorMessage msg = m.key();
+        QString fileName = m.key().fileName();
+        QFileInfo fi (fileName);
+        if (fi.isRelative()) {
+            fileName = oldPath.absoluteFilePath(fileName);
+        }
+        
+        msg.setFileName(fileName);
+        newmm.insert(msg, m.value());
+    }
+    mm = newmm;
+}
+
 void MetaTranslator::setCodec( const char *name )
 {
     const int latin1 = 4;
@@ -593,3 +731,38 @@ QList<MetaTranslatorMessage> MetaTranslator::translatedMessages() const
     }
     return val;
 }
+
+// the grammatical numerus is the number of plural forms + singular forms.
+// i.e english has two forms: singular og plural.
+// and polish has three forms:
+// 1. singular (1),
+// 2. plural form 1 (numbers that ends with 2,3,4 except 12,13,14)
+// 3. plural form 2 (all others)
+// Thus, english returns 2, polish returns 3
+int MetaTranslator::grammaticalNumerus(QLocale::Language language, QLocale::Country country)
+{
+    QStringList forms;
+    getNumerusInfo(language, country, &forms);
+    return forms.count();
+}
+
+QStringList MetaTranslator::normalizedTranslations(const MetaTranslatorMessage& m,
+                                                   QLocale::Language language,
+                                                   QLocale::Country country)
+{
+    QStringList translations = m.translations();
+    int numTranslations = 1;
+    if (m.isPlural()) {
+        numTranslations = grammaticalNumerus(language, country);
+    }
+
+    // make sure that the stringlist always have the size of the language's current numerus, or 1 if its not plural
+    if (translations.count() > numTranslations) {
+        for (int i = translations.count(); i >  numTranslations; --i)
+            translations.removeLast();
+    } else if (translations.count() < numTranslations) {
+        for (int i = translations.count(); i < numTranslations; ++i)
+            translations << QString();
+    }
+    return translations;
+}
