@@ -1,6 +1,6 @@
 // This is the implementation of the Chimera class.
 //
-// Copyright (c) 2010 Riverbank Computing Limited <info@riverbankcomputing.com>
+// Copyright (c) 2011 Riverbank Computing Limited <info@riverbankcomputing.com>
 // 
 // This file is part of PyQt.
 // 
@@ -38,6 +38,7 @@
 #include "qpycore_chimera.h"
 #include "qpycore_misc.h"
 #include "qpycore_pyqtpyobject.h"
+#include "qpycore_qpynullvariant.h"
 #include "qpycore_sip.h"
 #include "qpycore_types.h"
 
@@ -58,6 +59,21 @@ Chimera::Chimera()
 }
 
 
+// Construct a copy.
+Chimera::Chimera(const Chimera &other)
+{
+    _type = other._type;
+
+    _py_type = other._py_type;
+    Py_XINCREF(_py_type);
+    
+    _metatype = other._metatype;
+    _inexact = other._inexact;
+    _is_flag = other._is_flag;
+    _name = other._name;
+}
+
+
 // Destroy the type.
 Chimera::~Chimera()
 {
@@ -116,6 +132,21 @@ const Chimera *Chimera::parse(PyObject *obj)
 }
 
 
+// Parse a C++ type name as a type.
+const Chimera *Chimera::parse(const QByteArray &name)
+{
+    Chimera *ct = new Chimera;
+
+    if (!ct->parse_cpp_type(name))
+    {
+        delete ct;
+        return 0;
+    }
+
+    return ct;
+}
+
+
 // Parse a meta-property as a type.
 const Chimera *Chimera::parse(const QMetaProperty &mprop)
 {
@@ -284,17 +315,32 @@ Chimera::Signature *Chimera::parse(PyObject *types, const char *name,
 }
 
 
-// Raise an exception after parse() has failed.
-void Chimera::raiseParseException(PyObject *obj, const char *context)
+// Raise an exception after parse() of a Python type has failed.
+void Chimera::raiseParseException(PyObject *type, const char *context)
 {
-    raiseParseException(Py_TYPE(obj)->tp_name, context);
+    if (PyType_Check(type))
+    {
+        PyErr_Format(PyExc_TypeError,
+                "Python type '%s' is not supported as %s type",
+                ((PyTypeObject *)type)->tp_name, context);
+    }
+    else
+    {
+        const char *cpp_type_name = sipString_AsASCIIString(&type);
+
+        if (cpp_type_name)
+        {
+            raiseParseException(cpp_type_name, context);
+            Py_DECREF(type);
+        }
+    }
 }
 
 
-// Raise an exception after parse() has failed.
+// Raise an exception after parse() of a C++ type has failed.
 void Chimera::raiseParseException(const char *type, const char *context)
 {
-    PyErr_Format(PyExc_TypeError, "type '%s' is not supported as %s type",
+    PyErr_Format(PyExc_TypeError, "C++ type '%s' is not supported as %s type",
             type, context);
 }
 
@@ -302,9 +348,6 @@ void Chimera::raiseParseException(const char *type, const char *context)
 // Parse the given Python type object.
 bool Chimera::parse_py_type(PyTypeObject *type_obj)
 {
-    _py_type = (PyObject *)type_obj;
-    Py_INCREF(_py_type);
-
     const sipTypeDef *td = sipTypeFromPyTypeObject(type_obj);
 
     if (td)
@@ -324,6 +367,8 @@ bool Chimera::parse_py_type(PyTypeObject *type_obj)
         }
         else
         {
+            bool is_a_QObject = PyType_IsSubtype(type_obj, sipTypeAsPyTypeObject(sipType_QObject));
+
             // If there is no assignment helper then assume it is a
             // pointer-type.
             if (!get_assign_helper())
@@ -331,6 +376,51 @@ bool Chimera::parse_py_type(PyTypeObject *type_obj)
 
             _metatype = QMetaType::type(_name.constData());
 
+            // This will deal with cases where the registered type is a
+            // super-class and specifically allows for multiple inheritance.
+            // The problem we are solving is that we want a QGraphicsWidget to
+            // be handled as a QGraphicsItem (which Qt registers) rather than a
+            // QObject, specifically in the value passed as a QVariant to
+            // QGraphicsItem::itemChange().  This solution means that it will
+            // always be passed as a QGraphicsItem, even if there are
+            // circumstances where it should be passed as a QObject.  If we
+            // come across such a circumstance then we may need to remove this
+            // and implement a more specific solution in %MethodCode for the
+            // various itemChange() implementations.
+            if (_metatype == 0 && is_a_QObject)
+            {
+                PyObject *mro = type_obj->tp_mro;
+
+                Q_ASSERT(PyTuple_Check(mro));
+                Q_ASSERT(PyTuple_GET_SIZE(mro) >= 3);
+
+                for (SIP_SSIZE_T i = 1; i < PyTuple_GET_SIZE(mro) - 1; ++i)
+                {
+                    PyTypeObject *sc = (PyTypeObject *)PyTuple_GET_ITEM(mro, i);
+
+                    if (sc == sipSimpleWrapper_Type || sc == sipWrapper_Type)
+                        continue;
+
+                    QByteArray sc_name(sc->tp_name);
+
+                    // QObjects are always pointers.
+                    sc_name.append('*');
+
+                    _metatype = QMetaType::type(sc_name.constData());
+
+                    if (_metatype >= QMetaType::User)
+                    {
+                        _type = sipTypeFromPyTypeObject(sc);
+                        _name = sc_name;
+
+                        _py_type = (PyObject *)sc;
+                        Py_INCREF(_py_type);
+
+                        return true;
+                    }
+                }
+            }
+
             // If it is a user type then it must be a type that SIP knows
             // about but was registered by Qt.
             if (_metatype < QMetaType::User)
@@ -339,7 +429,7 @@ bool Chimera::parse_py_type(PyTypeObject *type_obj)
                 {
                     _metatype = QMetaType::QWidgetStar;
                 }
-                else if (PyType_IsSubtype(type_obj, sipTypeAsPyTypeObject(sipType_QObject)))
+                else if (is_a_QObject)
                 {
                     _metatype = QMetaType::QObjectStar;
                 }
@@ -408,6 +498,9 @@ bool Chimera::parse_py_type(PyTypeObject *type_obj)
     if (_name.isEmpty())
         _name = QMetaType::typeName(_metatype);
 
+    _py_type = (PyObject *)type_obj;
+    Py_INCREF(_py_type);
+
     return true;
 }
 
@@ -420,51 +513,156 @@ void Chimera::set_flag()
 }
 
 
+// Update a C++ type so that any typedefs and registered ints are resolved.
+QByteArray Chimera::resolve_types(const QByteArray &type)
+{
+    // Split into a base type and a possible list of template arguments.
+    QByteArray resolved = type.simplified();
+
+    // Get the raw type, ie. without any "const", "&" or "*".
+    QByteArray raw_type;
+    int original_raw_start;
+
+    if (resolved.startsWith("const "))
+        original_raw_start = 6;
+    else
+        original_raw_start = 0;
+
+    raw_type = resolved.mid(original_raw_start);
+
+    while (raw_type.endsWith('&') || raw_type.endsWith('*') || raw_type.endsWith(' '))
+        raw_type.chop(1);
+
+    int original_raw_len = raw_type.size();
+
+    if (original_raw_len == 0)
+        return QByteArray();
+
+    // Get any template arguments.
+    QList<QByteArray> args;
+    int tstart = raw_type.indexOf('<');
+
+    if (tstart >= 0)
+    {
+        // Make sure the template arguments are terminated.
+        if (!raw_type.endsWith('>'))
+            return QByteArray();
+
+        // Split the template arguments taking nested templates into account.
+        int depth = 1, arg_start = tstart + 1;
+
+        for (int i = arg_start; i < raw_type.size(); ++i)
+        {
+            int arg_end = -1;
+            char ch = raw_type.at(i);
+
+            if (ch == '<')
+            {
+                ++depth;
+            }
+            else if (ch == '>')
+            {
+                --depth;
+
+                if (depth < 0)
+                    return QByteArray();
+
+                if (depth == 0)
+                    arg_end = i;
+            }
+            else if (ch == ',' && depth == 1)
+            {
+                arg_end = i;
+            }
+
+            if (arg_end >= 0)
+            {
+                QByteArray arg = resolve_types(raw_type.mid(arg_start, arg_end - arg_start));
+
+                if (arg.isEmpty())
+                    return QByteArray();
+
+                args.append(arg);
+
+                arg_start = arg_end + 1;
+            }
+        }
+
+        if (depth != 0)
+            return QByteArray();
+
+        // Remove the template arguments.
+        raw_type.truncate(tstart);
+    }
+
+    // Expand any typedef.
+    const char *base_type = sipResolveTypedef(raw_type.constData());
+
+    if (base_type)
+        raw_type = base_type;
+
+    // Do the same for any registered int types.
+    if (_registered_int_types.contains(raw_type))
+        raw_type = "int";
+
+    // Add any (now resolved) template arguments.
+    if (args.count() > 0)
+    {
+        raw_type.append('<');
+
+        for (QList<QByteArray>::const_iterator it = args.begin();;)
+        {
+            raw_type.append(*it);
+
+            if (++it == args.end())
+                break;
+
+            raw_type.append(',');
+        }
+
+        if (raw_type.endsWith('>'))
+            raw_type.append(' ');
+
+        raw_type.append('>');
+    }
+
+    // Replace the original raw type with the resolved one.
+    resolved.replace(original_raw_start, original_raw_len, raw_type);
+
+    return resolved;
+}
+
+
 // Parse the given C++ type name.
 bool Chimera::parse_cpp_type(const QByteArray &type)
 {
-    // Get the expanded, normalised name.
-    QByteArray norm_name = _name = type;
+    _name = type;
 
-    // Extract the raw type and resolve any typedef.
-    QByteArray raw_type;
-    int raw_start;
-    const char *base_type;
+    // Resolve any types.
+    QByteArray resolved = resolve_types(type);
 
-    raw_start = extract_raw_type(norm_name, raw_type);
-    base_type = sipResolveTypedef(raw_type.constData());
-
-    if (base_type)
-        norm_name.replace(raw_start, raw_type.size(), base_type);
-
-    // Do the same for any registered int types.
-    if (!_registered_int_types.isEmpty())
-    {
-        raw_start = extract_raw_type(norm_name, raw_type);
-
-        if (_registered_int_types.contains(raw_type))
-            norm_name.replace(raw_start, raw_type.size(), "int");
-    }
+    if (resolved.isEmpty())
+        return false;
 
     // See if the type is known to Qt.
-    _metatype = QMetaType::type(norm_name.constData());
+    _metatype = QMetaType::type(resolved.constData());
 
     // If not then use the PyQt_PyObject wrapper.
     if (_metatype == QMetaType::Void)
         _metatype = PyQt_PyObject::metatype;
 
     // See if the type (without a pointer) is known to SIP.
-    bool is_ptr = norm_name.endsWith('*');
+    bool is_ptr = resolved.endsWith('*');
 
     if (is_ptr)
     {
-        norm_name.chop(1);
+        resolved.chop(1);
 
-        if (norm_name.endsWith('*'))
+        if (resolved.endsWith('*'))
             return false;
     }
 
-    _type = sipFindType(norm_name.constData());
+    _type = sipFindType(resolved.constData());
 
     if (!_type)
     {
@@ -476,7 +674,7 @@ bool Chimera::parse_cpp_type(const QByteArray &type)
         if (_metatype != PyQt_PyObject::metatype && !is_ptr)
             return true;
 
-        if ((norm_name == "char" || norm_name == "const char") && is_ptr)
+        if ((resolved == "char" || resolved == "const char") && is_ptr)
         {
             // This is a special value meaning a (hopefully) '\0' terminated
             // string.
@@ -486,7 +684,7 @@ bool Chimera::parse_cpp_type(const QByteArray &type)
         }
 
         // This is an explicit 'PyQt_PyObject'.
-        if (norm_name == "PyQt_PyObject" && !is_ptr)
+        if (resolved == "PyQt_PyObject" && !is_ptr)
             return true;
 
         return false;
@@ -525,37 +723,6 @@ bool Chimera::parse_cpp_type(const QByteArray &type)
 }
 
 
-// Extract the raw type from a normalised type name and return the start
-// position of the raw type in the original.
-int Chimera::extract_raw_type(const QByteArray &type, QByteArray &raw_type)
-{
-    // Find the start of the raw type.
-    int raw_start;
-
-    if (type.startsWith("const "))
-        raw_start = 6;
-    else
-        raw_start = 0;
-
-    // Find the length of the raw type, ie. before the start of any trailing
-    // indirection or reference.
-    int raw_len;
-
-    for (raw_len = type.size() - raw_start; raw_len >= 0; --raw_len)
-    {
-        char ch = type.at(raw_start + raw_len - 1);
-
-        if (ch != '&' && ch != '*')
-            break;
-    }
-
-    // Extract the raw type.
-    raw_type = type.mid(raw_start, raw_len);
-
-    return raw_start;
-}
-
-
 // Convert a Python object to C++ at a given address.  This has a lot in common
 // with the method that converts to a QVariant.  However, unlike that method,
 // we have no control over the size of the destination storage and so must
@@ -657,12 +824,21 @@ bool Chimera::fromPyObject(PyObject *py, void *cpp) const
 
     case QMetaType::QVariantList:
         {
-            QVariantList ql;
+            QList<QObject *> qlo;
 
-            if (to_QVariantList(py, ql))
-                *reinterpret_cast<QVariantList *>(cpp) = ql;
+            if (to_QList_QObject(py, qlo))
+            {
+                *reinterpret_cast<QList<QObject *> *>(cpp) = qlo;
+            }
             else
-                iserr = 1;
+            {
+                QVariantList ql;
+
+                if (to_QVariantList(py, ql))
+                    *reinterpret_cast<QVariantList *>(cpp) = ql;
+                else
+                    iserr = 1;
+            }
 
             break;
         }
@@ -960,16 +1136,26 @@ bool Chimera::fromPyObject(PyObject *py, QVariant *var, bool strict) const
 
     case QMetaType::QVariantList:
         {
-            QVariantList ql;
+            QList<QObject *> qlo;
 
-            if (to_QVariantList(py, ql))
+            if (to_QList_QObject(py, qlo))
             {
-                *var = QVariant(ql);
+                *var = QVariant(QList_QObject_metatype(), &qlo);
                 metatype_used = QMetaType::Void;
             }
             else
             {
-                iserr = 1;
+                QVariantList ql;
+
+                if (to_QVariantList(py, ql))
+                {
+                    *var = QVariant(ql);
+                    metatype_used = QMetaType::Void;
+                }
+                else
+                {
+                    iserr = 1;
+                }
             }
 
             break;
@@ -1127,7 +1313,7 @@ PyObject *Chimera::toPyObject(const QVariant &var) const
         {
             PyErr_Format(PyExc_TypeError,
                     "unable to convert a QVariant of type %d to a QMetaType of type %d",
-                    _metatype, var.userType());
+                    var.userType(), _metatype);
 
             return 0;
         }
@@ -1152,6 +1338,34 @@ PyObject *Chimera::toPyObject(const QVariant &var) const
         }
     }
 
+    // Handle QList<QObject *>.
+    int lo_metatype = QList_QObject_metatype();
+
+    if (lo_metatype != 0 && lo_metatype == var.userType())
+    {
+        const QList<QObject *> *cpp = reinterpret_cast<const QList<QObject *> *>(var.data());
+
+        PyObject *py_list = PyList_New(cpp->count());
+
+        if (!py_list)
+            return 0;
+
+        for (int i = 0; i < cpp->count(); ++i)
+        {
+            PyObject *obj = sipConvertFromType(cpp->at(i), sipType_QObject, 0);
+
+            if (!obj)
+            {
+                Py_DECREF(py_list);
+                return 0;
+            }
+
+            PyList_SET_ITEM(py_list, i, obj);
+        }
+
+        return py_list;
+    }
+
     return toPyObject(const_cast<void *>(var.data()));
 }
 
@@ -1425,6 +1639,10 @@ PyObject *Chimera::toAnyPyObject(const QVariant &var)
         return Py_None;
     }
 
+    // Handle null variants for the v2 API.
+    if (var.isNull() && sipIsAPIEnabled("QVariant", 2, 0))
+        return qpycore_qpynullvariant_FromQVariant(var);
+
     const char *type_name = var.typeName();
     const sipTypeDef *td = sipFindType(type_name);
 
@@ -1453,6 +1671,52 @@ QVariant Chimera::keep_as_pyobject(PyObject *py)
 }
 
 
+// Get the metatype for QList<QObject *> or 0 if it hasn't been registered.
+int Chimera::QList_QObject_metatype()
+{
+    static int lo_metatype = 0;
+
+    // We look each time until we find it because QtDeclarative could register
+    // it at any time.
+    if (lo_metatype == 0)
+        lo_metatype = QMetaType::type("QList<QObject*>");
+
+    return lo_metatype;
+}
+
+
+// Convert a Python list object to a QList<QObject*> and return true if there
+// was no error.
+bool Chimera::to_QList_QObject(PyObject *py, QList<QObject *>&cpp) const
+{
+    // We only try to convert if QList<QObject *> is a registered metatype.
+    if (QList_QObject_metatype() == 0)
+        return false;
+
+    Q_ASSERT(PyList_CheckExact(py));
+
+    for (SIP_SSIZE_T i = 0; i < PyList_GET_SIZE(py); ++i)
+    {
+        PyObject *val_obj = PyList_GET_ITEM(py, i);
+
+        if (!val_obj)
+            return false;
+
+        int iserr = 0;
+
+        QObject *val = reinterpret_cast<QObject *>(sipForceConvertToType(
+                val_obj, sipType_QObject, 0, SIP_NO_CONVERTORS, 0, &iserr));
+
+        if (iserr)
+            return false;
+
+        cpp.append(val);
+    }
+
+    return true;
+}
+
+
 // Convert a Python list object to a QVariantList and return true if there was
 // no error.
 bool Chimera::to_QVariantList(PyObject *py, QVariantList &cpp) const
@@ -1469,7 +1733,7 @@ bool Chimera::to_QVariantList(PyObject *py, QVariantList &cpp) const
         int val_state, iserr = 0;
 
         QVariant *val = reinterpret_cast<QVariant *>(sipForceConvertToType(
-                val_obj, sipType_QVariant, NULL, SIP_NOT_NONE, &val_state,
+                val_obj, sipType_QVariant, 0, SIP_NOT_NONE, &val_state,
                 &iserr));
 
         if (iserr)
