@@ -1,6 +1,6 @@
 // This is the interface of the Chimera and related classes.
 //
-// Copyright (c) 2010 Riverbank Computing Limited <info@riverbankcomputing.com>
+// Copyright (c) 2011 Riverbank Computing Limited <info@riverbankcomputing.com>
 // 
 // This file is part of PyQt.
 // 
@@ -47,6 +47,9 @@
 class Chimera
 {
 public:
+    // Construct a copy.
+    Chimera(const Chimera &other);
+
     // Destroy the type.
     ~Chimera();
 
@@ -56,6 +59,9 @@ public:
     // Parses a Qt meta-property as a type.  Return 0 if there was an error.
     static const Chimera *parse(const QMetaProperty &mprop);
 
+    // Parses a C++ type name as a type.  Return 0 if there was an error.
+    static const Chimera *parse(const QByteArray &name);
+
     class Signature
     {
     public:
@@ -72,6 +78,9 @@ public:
         // but excluding a return type).
         QByteArray signature;
 
+        // The signature as declared by the user for use in exceptions.
+        QByteArray py_signature;
+
         // The optional docstring which will start with '\1' if it is auto
         // generated and so can be used in exceptions.
         const char *docstring;
@@ -103,7 +112,7 @@ public:
         bool _cached;
 
         Signature(const QByteArray &sig, bool cached)
-            : result(0), signature(sig), docstring(0), _cached(cached) {}
+            : result(0), signature(sig), py_signature(sig), docstring(0), _cached(cached) {}
 
         Signature(const Signature &);
         Signature &operator=(const Signature &);
@@ -118,7 +127,7 @@ public:
             const char *context);
 
     // Raise an exception after parse() has failed.
-    static void raiseParseException(PyObject *obj, const char *context);
+    static void raiseParseException(PyObject *type, const char *context);
 
     class Storage
     {
@@ -179,7 +188,7 @@ public:
 
     // Convert a Python object to a QVariant.  Return false if there was an
     // error.
-    bool fromPyObject(PyObject *py, QVariant *var) const;
+    bool fromPyObject(PyObject *py, QVariant *var, bool strict = true) const;
 
     // Convert a Python object to a QVariant based on the type of the object.
     static QVariant fromAnyPyObject(PyObject *py, int *is_err);
@@ -197,6 +206,10 @@ public:
     // isn't known to Qt's meta-type system.
     int metatype() const {return _metatype;}
 
+    // Returns a borrowed reference to the Python type object that was used to
+    // define the type (if any).
+    PyObject *py_type() const {return _py_type;}
+
     // Returns the C++ name of the type.
     const QByteArray &name() const {return _name;}
 
@@ -217,6 +230,10 @@ private:
     // but hasn't been wrapped.
     const sipTypeDef *_type;
 
+    // The Python type object if a Python type (rather than a C++ type string)
+    // was given.
+    PyObject *_py_type;
+
     // The Qt meta-type.  This will always be valid but may refer to the
     // PyObject wrapper if the underlying type isn't known to Qt.
     int _metatype;
@@ -242,6 +259,8 @@ private:
     bool parse_py_type(PyTypeObject *type_obj);
     sipAssignFunc get_assign_helper() const;
     void set_flag();
+    static int QList_QObject_metatype();
+    bool to_QList_QObject(PyObject *py, QList<QObject *> &cpp) const;
     bool to_QVariantList(PyObject *py, QVariantList &cpp) const;
     bool to_QVariantMap(PyObject *py, QVariantMap &cpp) const;
 #if QT_VERSION >= 0x040500
@@ -252,8 +271,8 @@ private:
     static void raiseParseException(const char *type, const char *context);
     static QVariant keep_as_pyobject(PyObject *py);
     static int extract_raw_type(const QByteArray &type, QByteArray &raw_type);
+    static QByteArray resolve_types(const QByteArray &type);
 
-    Chimera(const Chimera &);
     Chimera &operator=(const Chimera &);
 };
 
