@@ -1,6 +1,6 @@
 // This defines the interfaces for the pyqtSignal type.
 //
-// Copyright (c) 2010 Riverbank Computing Limited <info@riverbankcomputing.com>
+// Copyright (c) 2011 Riverbank Computing Limited <info@riverbankcomputing.com>
 // 
 // This file is part of PyQt.
 // 
@@ -42,14 +42,20 @@
 extern "C" {
 
 // This defines the structure of a PyQt signal.
-typedef struct {
+typedef struct _qpycore_pyqtSignal {
     PyObject_HEAD
 
+    // The signal master.
+    struct _qpycore_pyqtSignal *master;
+
+    // The index in the signal master.
+    int master_index;
+
     // The non-signal overloads (if any).
     PyMethodDef *non_signals;
 
     // The list of overloaded signals.  (A pointer because C is creating the
-    // instance.)
+    // instance.)  These are owned by the signal master.
     QList<Chimera::Signature *> *overloads;
 } qpycore_pyqtSignal;
 
@@ -62,6 +68,11 @@ int qpycore_get_lazy_attr(const sipTypeDef *td, PyObject *dict);
 }
 
 
+qpycore_pyqtSignal *qpycore_pyqtSignal_New();
+void qpycore_add_native_Qt_signal(qpycore_pyqtSignal *ps, const char *sig,
+        const char *docstring);
+int qpycore_signal_index(qpycore_pyqtSignal *ps, PyObject *subscript,
+        const char *context);
 void qpycore_set_signal_name(qpycore_pyqtSignal *ps, const char *type_name,
         const char *name);
 PyObject *qpycore_call_signal_overload(PyObject *ps_obj, PyObject *bound,
