@@ -1,6 +1,6 @@
 // This is the signal/slot helper code for SIP.
 //
-// Copyright (c) 2010 Riverbank Computing Limited <info@riverbankcomputing.com>
+// Copyright (c) 2011 Riverbank Computing Limited <info@riverbankcomputing.com>
 // 
 // This file is part of PyQt.
 // 
@@ -38,6 +38,7 @@
 #include <QObject>
 
 #include "qpycore_chimera.h"
+#include "qpycore_pyqtboundsignal.h"
 #include "qpycore_pyqtproxy.h"
 #include "qpycore_pyqtpyobject.h"
 #include "qpycore_sip.h"
@@ -147,7 +148,8 @@ extern "C" void *sipQtCreateUniversalSignal(void *tx, const char **sigp)
 
 
 // Factory function to create a universal slot instance.  Returns a pointer to
-// the instance or 0 if there was an error.
+// the instance or 0 if there was an error.  Note that we may also return a
+// Qt signal (from a bound signal).
 extern "C" void *sipQtCreateUniversalSlot(sipWrapper *tx, const char *sig,
         PyObject *rxObj, const char *slot, const char **member, int flags)
 {
@@ -155,11 +157,25 @@ extern "C" void *sipQtCreateUniversalSlot(sipWrapper *tx, const char *sig,
     PyObject *qrxObj;
 
     if (slot)
+    {
         qrxObj = rxObj;
+    }
+    else if (Py_TYPE(rxObj) == &qpycore_pyqtBoundSignal_Type)
+    {
+        qpycore_pyqtBoundSignal *bs = (qpycore_pyqtBoundSignal *)rxObj;
+
+        *member = bs->bound_overload->signature.constData();
+
+        return bs->bound_qobject;
+    }
     else if (PyMethod_Check(rxObj))
+    {
         qrxObj = PyMethod_GET_SELF(rxObj);
+    }
     else
+    {
         qrxObj = 0;
+    }
 
     QObject *qrx = 0;
 
