@@ -1,6 +1,6 @@
 // This defines access to the SIP helpers for the rest of the library.
 //
-// Copyright (c) 2010 Riverbank Computing Limited <info@riverbankcomputing.com>
+// Copyright (c) 2011 Riverbank Computing Limited <info@riverbankcomputing.com>
 // 
 // This file is part of PyQt.
 // 
@@ -37,7 +37,9 @@
 #include "qpycore_chimera.h"
 
 
+QT_BEGIN_NAMESPACE
 class QObject;
+QT_END_NAMESPACE
 
 
 // Forward declarations.
